// -*-c++-*-

/*!
  \file domain_graph.h
  \brief data structures and procedures to populate and search in domain graphs
*/

/*
 *Copyright:

 Copyright (C) Majd Hawasly

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_DOMAIN_GRAPH_H
#define RCSC_DOMAIN_GRAPH_H

#include "action_generator.h"
#include "field_evaluator.h"
#include "sample_communication.h"

#include <rcsc/geom/vector_2d.h>
#include "rcsc/player/intercept_table.h"
#include "cooperative_action.h"

#include <boost/shared_ptr.hpp>

#include <vector>
#include <iostream>
#include <utility>
#include <cstdio>

namespace rcsc {
class PlayerAgent;
class WorldModel;
}

class CooperativeAction;
class PredictState;

class DomainGraph {
public:

    typedef boost::shared_ptr< DomainGraph > Ptr; //!< pointer type alias
    typedef boost::shared_ptr< const DomainGraph > ConstPtr; //!< const pointer type alias

public:
    static const size_t NUM_TEAM_STATES=14;	//0:Unknown, 1-11:Hold, 12:Goal, 13:Lost
    static const size_t NUM_LOCAL_GAMES=6;//8;	//0:Hold, 1:Dribble, 2:Pass, 3:Shoot, 4:clear (not used), 5:getBall (a.k.a. Move)
    static const size_t NUM_FEATURES=3;
    static const double ROBUSTNESS_THRESHOLD=0.05;
    static const int POSS_THRESHOLD=8;	//number of cycles to ball possession
    static const double ALPHA=0.33;
    static const double MAX_BALL_RANGE=40.0;


private:
    FieldEvaluator::ConstPtr M_evaluator;
    ActionGenerator::ConstPtr M_action_generator;
    
    CooperativeAction::ConstPtr M_action;

    
    double V[NUM_TEAM_STATES];
    double R[NUM_TEAM_STATES][NUM_LOCAL_GAMES][NUM_TEAM_STATES][NUM_FEATURES];
    //static const std::string robustness_dir;;
   
private:
    std::vector< ActionStatePair > M_result;
    double M_best_evaluation;
    int M_current_Tstate;
    int M_old_Tstate;
    CooperativeAction::ActionCategory chosen_link;
    CooperativeAction::ActionCategory unlearnt_link;
    int unlearnt_link_src;
    int unlearnt_link_pars;
    double heardTstates[NUM_TEAM_STATES];
    bool lossTstates[NUM_TEAM_STATES];
    bool goalTstates[NUM_TEAM_STATES];
    bool ownedTstates[NUM_TEAM_STATES];
    int features[NUM_LOCAL_GAMES];	//structure to keep computed features for reuse
    int f_times[NUM_LOCAL_GAMES];	//time when the feature was last computed
    int best_link;
    int lScore;	// variable to flag the goal event and activate the Tstate
    int rScore;	// variable to flag the goal event and activate the Tstate
    
    bool readRobustness(int role);
    bool writeRobustness(int role);
    void learnRobustness( const rcsc::WorldModel& wm);
    void learnRobustness(CooperativeAction::ActionCategory link, int mate_unum, const rcsc::WorldModel& wm);
    void learnRobustness(CooperativeAction::ActionCategory link, int mate_unum, const rcsc::WorldModel& wm, int src_Tstate);
    void updateValues(const rcsc::WorldModel& wm, int (&Values)[12]);
    void gameOutcome(int s,int g,const rcsc::WorldModel& wm, int toPassTo,double& max_v, int& blink);
    
    int role_num;


    int feature_hd(const rcsc::WorldModel& wm);
    int feature_p(const rcsc::WorldModel& wm);
    int feature_p(const rcsc::WorldModel& wm, int mate_unum);
    int feature_sh(const rcsc::WorldModel& wm);
    int feature_gb(const rcsc::WorldModel & wm);
    



public:

    
    
    DomainGraph( );

    static DomainGraph & instance();
    ValueMessage* valueMsg;
    bool loaded_Robustness;
    void loadRobustness();

    void findBestLink(const rcsc::WorldModel& wm, int toPassTo);
    
    int getBestLink()const{return best_link;}
    
   // CooperativeAction & action() const {return M_action;}
    const CooperativeAction::ConstPtr action() const
    {
	
	return M_action;//*
	
    }
   
    void setAction(CooperativeAction::ConstPtr action);
    
    
    bool active(int index) const	//check if the player was active so it took a decision
    {
        return (M_old_Tstate==index || M_old_Tstate==13);
    }

    int extractFeatures(int game, const rcsc::WorldModel & wm);
    int extractFeatures(int game, const rcsc::WorldModel & wm, int teammate_unum);
    

    int currentTstate() const
    {
        return M_current_Tstate;
    }

    void setChosenLink(CooperativeAction::ActionCategory cat)
    {
        chosen_link=cat;
    };
    
    void setRoleNum(int role)
    {
	role_num=role;
    }

    void update( const rcsc::WorldModel& wm, int (& Values)[12]);
};


#endif






#ifndef RCSC_PREDICATE_COMPLEX_H
#define RCSC_PREDICATE_COMPLEX_H

//#include "action_generator.h"
//#include "field_evaluator.h"
//#include "sample_communication.h"

#include <rcsc/geom/vector_2d.h>
//#include "rcsc/player/intercept_table.h"
//#include "cooperative_action.h"

#include <boost/shared_ptr.hpp>

#include <vector>
//#include <iostream>
//#include <utility>
//#include <cstdio>
class PredicateComplex {
public:

    typedef boost::shared_ptr< PredicateComplex > Ptr; //!< pointer type alias
    typedef boost::shared_ptr< const PredicateComplex > ConstPtr; //!< const pointer type alias

public:
    enum predicate
    {
	none,		//0
	open,		//1
	contact,	//2
	position,	//3
	ball	,	//4
	blocked	,	//5
	marked,		//6
	supported,	//7
	tacklable	//8
    };
    private:    CooperativeAction::ConstPtr  M_action;

public:
    PredicateComplex();
    static PredicateComplex & instance();
    void update(const rcsc::WorldModel& wm, int (& Marked)[12]);
    void generateTargets(const rcsc::WorldModel& wm, std::list< rcsc::Vector2D >& targets,bool offence);
    //void generateTargets(const rcsc::WorldModel& wm, std::list< rcsc::Vector2D >& targets, bool offence, PredicateComplex::predicate extra);
    std::vector<predicate> o_activePredicates;
    std::vector<predicate> d_activePredicates;
    bool isPosession(rcsc::Vector2D t, const rcsc::WorldModel& wm);
    bool isOpen(rcsc::Vector2D t, const rcsc::WorldModel& wm);
    bool isContact(rcsc::Vector2D,const rcsc::WorldModel& wm );
    bool isPosition(rcsc::Vector2D,const rcsc::WorldModel& wm );
    bool isBall(rcsc::Vector2D);
    bool isBlocked(rcsc::Vector2D,const rcsc::WorldModel& wm );
    bool isMarked(rcsc::Vector2D,const rcsc::WorldModel& wm );
    bool isSupported(rcsc::Vector2D,const rcsc::WorldModel& wm );
    bool isTacklable(rcsc::Vector2D,const rcsc::WorldModel& wm );
    MarkingMessage* markingMsg;
    int marked_unum;
    bool blocking;
    void setAction(CooperativeAction::ConstPtr action);

    bool Marked(int unum)
    {
    	//the ball and the 11 opponents
    	if(unum>=0 && unum<=11) return (M[unum]!=-1);
    	return true;
    }

    int Marker(int unum)
    {
    	//the ball and the 11 opponents
    	if(unum>=0 && unum<=11) return (M[unum]);
    	return -1;
    }

    const CooperativeAction::ConstPtr action() const
    {
    	return M_action;
    }

    int pass_mate;


private:
    int M[12];
    int MT[12];



};
#endif

